@@ -3,10 +3,10 @@
  *  Module    : mail.c
  *  Author    : I. Lea
  *  Created   : 1992-10-02
- *  Updated   : 2011-11-06
+ *  Updated   : 2016-07-29
  *  Notes     : Mail handling routines for creating pseudo newsgroups
  *
- * Copyright (c) 1992-2012 Iain Lea <iain@bricbrac.de>
+ * Copyright (c) 1992-2016 Iain Lea <iain@bricbrac.de>
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -207,11 +207,15 @@ write_mail_active_file(
 		}
 		if ((i = ferror(fp)) || fclose(fp)) {
 			error_message(2, _(txt_filesystem_full), mail_active_file);
-			rename(file_tmp, mail_active_file);
 			if (i) {
 				clearerr(fp);
 				fclose(fp);
 			}
+			i = rename(file_tmp, mail_active_file);
+#ifdef DEBUG
+			if ((debug & DEBUG_MISC) && i) /* TODO: is this the right debug-level? */
+				perror_message(_(txt_rename_error), file_tmp, mail_active_file);
+#endif /* DEBUG */
 		} else
 			unlink(file_tmp);
 	}
@@ -268,19 +272,19 @@ open_newsgroups_fp(
 
 	if (read_news_via_nntp && !read_saved_news) {
 		if (read_local_newsgroups_file) {
-			struct stat buf;
+			if ((result = fopen(local_newsgroups_file, "r")) != NULL) {
+				struct stat buf;
 
-			if (!stat(local_newsgroups_file, &buf)) {
-				if (buf.st_size > 0) {
-					if ((result = fopen(local_newsgroups_file, "r")) != NULL) {
 #	ifdef DEBUG
-						if (debug & DEBUG_NNTP)
-							debug_print_file("NNTP", "open_newsgroups_fp Using local copy of newsgroups file");
+					if (debug & DEBUG_NNTP)
+						debug_print_file("NNTP", "open_newsgroups_fp Using local copy of newsgroups file");
 #	endif /* DEBUG */
+				if (!fstat(fileno(result), &buf)) {
+					if (buf.st_size > 0)
 						return result;
-					}
-				} else
-					unlink(local_newsgroups_file);
+				}
+				fclose(result);
+				unlink(local_newsgroups_file);
 			}
 		}
 		/*
@@ -311,7 +315,7 @@ open_newsgroups_fp(
 						if (group->type == GROUP_TYPE_NEWS) {
 							if (nntp_caps.type == CAPABILITIES && nntp_caps.list_newsgroups) {
 								if (*buff) {
-									if (strlen(buff) + strlen(active[i].name) + 1 < NNTP_STRLEN) {
+									if (strlen(buff) + strlen(active[i].name) + 1 < NNTP_GRPLEN) {
 										snprintf(buff + strlen(buff), sizeof(buff) - strlen(buff), ",%s", active[i].name);
 										continue;
 									} else {
@@ -368,10 +372,8 @@ open_newsgroups_fp(
 				}
 				/* TODO: add 483 (RFC 3977) support */
 				if (no_more_wildmat == ERR_NOAUTH || no_more_wildmat == NEED_AUTHINFO) {
-					if (!authenticate(nntp_server, userid, FALSE)) {
-						error_message(2, _(txt_auth_failed), nntp_caps.type == CAPABILITIES ? ERR_AUTHFAIL : ERR_ACCESS);
-						tin_done(EXIT_FAILURE);
-					}
+					if (!authenticate(nntp_server, userid, FALSE))
+						tin_done(EXIT_FAILURE, _(txt_auth_failed), nntp_caps.type == CAPABILITIES ? ERR_AUTHFAIL : ERR_ACCESS);
 				}
 #		endif /* !DISABLE_PIPELINING */
 				fclose(result);
@@ -387,9 +389,9 @@ open_newsgroups_fp(
 			}
 		}
 		return (nntp_command("LIST NEWSGROUPS", OK_GROUPS, NULL, 0));
-	} else
+	}
 #endif /* NNTP_ABLE */
-		return fopen(newsgroups_file, "r");
+	return fopen(newsgroups_file, "r");
 }
 
 
@@ -408,12 +410,8 @@ read_newsgroups_file(
 		if (!batch_mode && verb)
 			wait_message(0, _(txt_reading_newsgroups_file));
 
-		if (read_news_via_nntp && !no_write) {
-			struct stat buf;
-
-			if (stat(local_newsgroups_file, &buf) || !read_local_newsgroups_file)
-				fp_save = fopen(local_newsgroups_file, "w");
-		}
+		if (read_news_via_nntp && !no_write && !read_local_newsgroups_file)
+			fp_save = fopen(local_newsgroups_file, "w");
 
 		read_groups_descriptions(fp, fp_save);
 
@@ -610,6 +608,7 @@ grp_del_mail_arts(
 		 * at least for GROUP_TYPE_SAVE a wait is annoying - nuke the message?
 		 */
 		wait_message(0, (group->type == GROUP_TYPE_MAIL) ? _(txt_processing_mail_arts) : _(txt_processing_saved_arts));
+		cursoroff();
 		make_base_group_path(group->spooldir, group->name, group_path, sizeof(group_path));
 		for_each_art(i) {
 			article = &arts[i];
@@ -648,7 +647,8 @@ art_edit(
 	make_base_group_path(group->spooldir, group->name, temp_filename, sizeof(temp_filename));
 	snprintf(buf, sizeof(buf), "%"T_ARTNUM_PFMT, article->artnum);
 	joinpath(article_filename, sizeof(article_filename), temp_filename, buf);
-	snprintf(temp_filename, sizeof(temp_filename), "%s%ld.art", TMPDIR, (long) process_id);
+	snprintf(buf, sizeof(buf), "%ld.art", (long) process_id);
+	joinpath(temp_filename, sizeof(temp_filename), TMPDIR, buf);
 
 	if (!backup_file(article_filename, temp_filename))
 		return FALSE;
@@ -656,8 +656,8 @@ art_edit(
 	if (!invoke_editor(temp_filename, 1, group)) {
 		unlink(temp_filename);
 		return FALSE;
-	} else
-		rename_file(temp_filename, article_filename);
+	}
 
+	rename_file(temp_filename, article_filename);
 	return TRUE;
 }
