@@ -3,7 +3,7 @@
  *  Module    : nntplib.c
  *  Author    : S. Barber & I. Lea
  *  Created   : 1991-01-12
- *  Updated   : 2011-09-12
+ *  Updated   : 2016-07-29
  *  Notes     : NNTP client routines taken from clientlib.c 1.5.11 (1991-02-10)
  *  Copyright : (c) Copyright 1991-99 by Stan Barber & Iain Lea
  *              Permission is hereby granted to copy, reproduce, redistribute
@@ -170,13 +170,7 @@ getserverbyfile(
 			(void) fclose(fp);
 			return buf;
 		}
-
 		(void) fclose(fp);
-
-		if (cp != NULL) {
-			get_nntpserver(buf, sizeof(buf), cp);
-			return buf;
-		}
 	}
 
 #	ifdef NNTP_DEFAULT_SERVER
@@ -260,12 +254,14 @@ server_init(
 #	ifdef TLI /* Transport Level Interface */
 	if (t_sync(sockt_rd) < 0) {	/* Sync up new fd with TLI */
 		t_error("server_init: t_sync()");
+		s_fclose(nntp_rd_fp);
 		nntp_rd_fp = NULL;
 		return -EPROTO;
 	}
 #	else
 	if ((nntp_wr_fp = (TCP *) s_fdopen(sockt_wr, "w")) == NULL) {
 		perror("server_init: fdopen() #2");
+		s_fclose(nntp_rd_fp);
 		nntp_rd_fp = NULL;
 		return -errno;
 	}
@@ -374,7 +370,7 @@ get_tcp_socket(
 	if (t_connect(s, callptr, (struct t_call *) 0) < 0) {
 		save_errno = t_errno;
 		if (save_errno == TLOOK)
-			fprintf(stderr, _(txt_error_server_unavailable));
+			fprintf(stderr, "%s", _(txt_error_server_unavailable));
 		else
 			t_error("t_connect");
 		t_free((char *) callptr, T_CALL);
@@ -601,7 +597,7 @@ get_tcp6_socket(
 	struct addrinfo hints, *res, *res0;
 
 	snprintf(mymachine, sizeof(mymachine), "%s", machine);
-	snprintf(myport, sizeof(myport), "%d", port);
+	snprintf(myport, sizeof(myport), "%u", port);
 
 /* just in case */
 #	ifdef AF_UNSPEC
@@ -776,17 +772,19 @@ put_server(
 		 * remember the last command we wrote to be able to resend it after a
 		 * reconnect. reconnection is handled by get_server()
 		 *
-		 * don't cache "LIST ACTIVE something" as we would need to
-		 * resend all of them but we remember just the last one. we cache
-		 * "LIST" instead, this will slow down things, but that's ok on
-		 * reconnect.
+		 * don't cache "LIST [ACTIVE|COUNTS|NEWSGROUPS] something" as we
+		 * would need to resend all of them but we remember just the last
+		 * one. we cache "LIST cmd." instead, this will slow down things, but
+		 * that's ok on reconnect.
 		 */
+        if (strcmp(last_put, string))
+        	STRCPY(last_put, string);
 		if (!strncmp(string, "LIST ACTIVE ", 12))
-			STRCPY(last_put, "LIST");
-		else {
-			if (last_put != string)
-				STRCPY(last_put, string);
-		}
+			last_put[11] = '\0'; /* "LIST ACTIVE" */
+		else if (!strncmp(string, "LIST COUNTS ", 12))
+			last_put[11] = '\0'; /* "LIST COUNTS" */
+		else if (!strncmp(string, "LIST NEWSGROUPS ", 16))
+			last_put[15] = '\0'; /* "LIST NEWSGROUPS" */
 	}
 	(void) s_flush(nntp_wr_fp);
 }
@@ -825,32 +823,35 @@ reconnect(
 	signal_context = cReconnect;
 
 	/*
-	 * Exit tin if the user says no to reconnect. The exit code stops tin from trying
-	 * to disconnect again - the connection is already dead
+	 * Exit tin if there are no more tries or if the user says no to reconnect.
+	 * The exit code stops tin from trying to disconnect again - the connection
+	 * is already dead
 	 */
-	if (!tinrc.auto_reconnect && prompt_yn(_(txt_reconnect_to_news_server), TRUE) != 1) {
-		if (!strncmp("POST", last_put, 4)) {
+	if (retry > NNTP_TRY_RECONNECT || (!tinrc.auto_reconnect && prompt_yn(_(txt_reconnect_to_news_server), TRUE) != 1)) {
+		if (!strcmp("POST", last_put)) {
 			unlink(backup_article_name(article_name));
 			rename_file(article_name, dead_article);
 			if (tinrc.keep_dead_articles)
-				append_file(dead_articles, dead_article);
+				append_file(dead_article, dead_articles);
 		}
-		tin_done(NNTP_ERROR_EXIT);		/* user said no to reconnect */
+		if (retry > NNTP_TRY_RECONNECT) {
+#	ifdef DEBUG
+			/* TODO: -> lang.c */
+			if (debug & DEBUG_NNTP)
+				debug_print_file("NNTP", "reconnect(%d) limit %d reached, giving up.", retry, NNTP_TRY_RECONNECT);
+#	endif /* DEBUG */
+		}
+		tin_done(NNTP_ERROR_EXIT, _("NNTP connection error. Exiting..."));		/* user said no to reconnect or no more retries */
 	}
 
-	/*
-	 * reset signal_context
-	 */
+	/* reset signal_context */
 	signal_context = save_signal_context;
 
 	clear_message();
-
 	strcpy(buf, last_put);			/* Keep copy here, it will be clobbered a lot otherwise */
 
 	if (!nntp_open()) {
-		/*
-		 * Re-establish our current group and resend last command
-		 */
+		/* Re-establish our current group and resend last command */
 		if (curr_group != NULL) {
 			DEBUG_IO((stderr, _("Rejoin current group\n")));
 			snprintf(last_put, sizeof(last_put), "GROUP %s", curr_group->name);
@@ -865,17 +866,7 @@ reconnect(
 		DEBUG_IO((stderr, _("Resend last command (%s)\n"), buf));
 		put_server(buf);
 		did_reconnect = TRUE;
-		return 0;
-	}
-
-	if (--retry == 0) {					/* No more tries? */
-		if (!strncmp("POST", buf, 4)) {
-			unlink(backup_article_name(article_name));
-			rename_file(article_name, dead_article);
-			if (tinrc.keep_dead_articles)
-				append_file(dead_articles, dead_article);
-		}
-		tin_done(NNTP_ERROR_EXIT);
+		retry = NNTP_TRY_RECONNECT;
 	}
 
 	return retry;
@@ -900,7 +891,7 @@ get_server(
 	char *string,
 	int size)
 {
-	int retry = NNTP_TRY_RECONNECT;
+	static int retry_cnt = 0;
 
 	reconnected_in_last_get_server = FALSE;
 	errno = 0;
@@ -928,7 +919,7 @@ get_server(
 		alarm(0);
 #	endif /* HAVE_ALARM && SIGALRM */
 		if (quitting)						/* Don't bother to reconnect */
-			tin_done(NNTP_ERROR_EXIT);		/* And don't try to disconnect again! */
+			tin_done(NNTP_ERROR_EXIT, NULL);		/* And don't try to disconnect again! */
 
 #	ifdef DEBUG
 		if (errno != 0 && errno != EINTR)	/* Will only confuse end users */
@@ -941,17 +932,17 @@ get_server(
 		 * closed immediately). Also prevents tin from asking to reconnect
 		 * when user is quitting tin if tinrc.auto_reconnect is false.
 		 */
-		if (strncmp(last_put, "QUIT", 4)) {
+		if (strcmp(last_put, "QUIT")) {
 			/*
 			 * Typhoon v2.1.1.363 colses the connection right after an unknown
 			 * command, (i.e. CAPABILITIES) so we avoid the reissue it on a
 			 * reconnect if it was the last command.
 			 */
-			if (!strncmp(last_put, "CAPABILITIES", 12)) {
+			if (!strcmp(last_put, "CAPABILITIES")) {
 				strcpy(last_put, "MODE READER");
 				nntp_caps.type = BROKEN;
 			}
-			retry = reconnect(retry);		/* Will abort when out of tries */
+			retry_cnt = reconnect(++retry_cnt);		/* Will abort when out of tries */
 			reconnected_in_last_get_server = TRUE;
 		} else {
 			/*
@@ -966,6 +957,7 @@ get_server(
 #	if defined(HAVE_ALARM) && defined(SIGALRM)
 	alarm(0);
 #	endif /* HAVE_ALARM && SIGALRM */
+	retry_cnt = 0;
 	return string;
 }
 
@@ -999,6 +991,7 @@ close_server(
 	nntp_wr_fp = nntp_rd_fp = NULL;
 }
 
+
 /*
  * Try and use CAPABILITIES here. Get this list before issuing other NNTP
  * commands because the correct methods may be mentioned in the list of
@@ -1007,7 +1000,8 @@ close_server(
  * Sets up: t_capabilities nntp_caps
  */
 static int
-check_extensions(void)
+check_extensions(
+	void)
 {
 	char *d;
 	char *ptr;
@@ -1036,7 +1030,7 @@ check_extensions(void)
 			nntp_caps.list_distributions = FALSE;
 			nntp_caps.list_moderators = FALSE;
 			nntp_caps.list_counts = FALSE;
-			nntp_caps.xpat = FALSE;
+			nntp_caps.xpat = TRUE; /* only used in select.c:lookup_msgid(); toggles to false if fails, INN > 2.7.0 announces it */
 			nntp_caps.hdr = FALSE;
 			nntp_caps.hdr_cmd = NULL;
 			nntp_caps.over = FALSE;
@@ -1205,7 +1199,7 @@ check_extensions(void)
 								nntp_caps.sasl |= SASL_LOGIN;
 							}
 						}
-					} else if (!strncasecmp(ptr, "COMPRESS", 8)) { /* draft-murchison-nntp-compress-01.txt */
+					} else if (!strncasecmp(ptr, "COMPRESS", 8)) { /* draft-murchison-nntp-compress-05 */
 						d = ptr + 8;
 						d = strpbrk(d, " \t");
 						while (d != NULL && (d + 1 < (ptr + strlen(ptr)))) {
@@ -1222,6 +1216,7 @@ check_extensions(void)
 					else if (!strcasecmp(ptr, "STREAMING"))
 						nntp_caps.streaming = TRUE;
 #		endif /* 0 */
+				/* XZVER, XZHDR, ... */
 				} else
 					nntp_caps.type = NONE;
 			}
@@ -1236,7 +1231,7 @@ check_extensions(void)
 #		if 1
 		case ERR_GOODBYE:
 			ret = i;
-			error_message(2, buf);
+			error_message(2, "%s", buf);
 			break;
 #		endif /* 1 */
 
@@ -1296,7 +1291,7 @@ mode_reader(
 
 			case ERR_GOODBYE:
 			case ERR_ACCESS:
-				error_message(2, line);
+				error_message(2, "%s", line);
 				return ret;
 
 			case ERR_COMMAND:
@@ -1400,7 +1395,7 @@ nntp_open(
 			if (ret < 0)
 				error_message(2, _(txt_failed_to_connect_to_server), nntp_server);
 			else
-				error_message(2, line);
+				error_message(2, "%s", line);
 
 			return ret;
 	}
@@ -1459,7 +1454,7 @@ nntp_open(
 				/* just honor ciritical errors */
 				case ERR_GOODBYE:
 				case ERR_ACCESS:
-					error_message(2, buf);
+					error_message(2, "%s", buf);
 					return -1;
 
 				default:
@@ -1573,7 +1568,6 @@ nntp_open(
 					break;
 			}
 		}
-#	ifdef XHDR_XREF
 		for (i = 0, j = 0; i < 2 && j >= 0; i++) {
 			j = new_nntp_command(&xhdr_cmds[i], ERR_CMDSYN, line, sizeof(line));
 			switch (j) {
@@ -1597,7 +1591,18 @@ nntp_open(
 					break;
 			}
 		}
-#	endif /* XHDR_XREF */
+		/* no XPAT probing here, we do it in select.c:lookup_msgid() as that's the only place we use it */
+		nntp_caps.xpat = TRUE;
+#		if 0
+		switch (new_nntp_command("XPAT Newsgroups <0> *", ERR_NOART, line, sizeof(line))) {
+			case ERR_NOART:
+				nntp_caps.xpat = TRUE;
+				break;
+
+			default:
+				break;
+		}
+#		endif /* 0 */
 	} else {
 		if (!nntp_caps.over_cmd) {
 			/*
@@ -1622,7 +1627,6 @@ nntp_open(
 					break;
 			}
 		}
-#	ifdef XHDR_XREF
 		if (!nntp_caps.hdr_cmd) {
 			/*
 			 * CAPABILITIES didn't mention HDR or XHDR, try XHDR
@@ -1646,7 +1650,6 @@ nntp_open(
 					break;
 			}
 		}
-#	endif /* XHDR_XREF */
 	}
 
 	if (!nntp_caps.over_cmd) {
@@ -1671,8 +1674,8 @@ nntp_open(
 	if (!is_reconnect && !batch_mode && show_description && check_for_new_newsgroups) {
 		/*
 		 * TODO:
-		 * - document that "-d" and/or "-q" turns off "LIST MOTD" (or add
-		 *   a tinrc var to turn LIST MOTD on/off)
+		 * - add a tinrc var to turn LIST MOTD on/off?
+		 *   (currently done automatically for -d, -q and -Q)
 		 */
 		if (nntp_caps.list_motd)
 			list_motd();
@@ -1740,13 +1743,17 @@ get_only_respcode(
 		debug_print_file("NNTP", "<<<%s%s", logtime(), ptr);
 #	endif /* DEBUG */
 	respcode = (int) strtol(ptr, &end, 10);
+	if (end == ptr) /* no leading numbers in response */
+		respcode = -1;
 	DEBUG_IO((stderr, "get_only_respcode(%d)\n", respcode));
 
 	/*
 	 * we also reconnect on ERR_FAULT if last_put was ARTICLE or LIST or POST
 	 * as inn (2.2.3) sends ERR_FAULT on timeout
+	 *
+	 * what about other LIST cmds? (ACTIVE|COUNTS|OVERVIEW.FMT|...)
 	 */
-	if (last_put[0] != '\0' && ((respcode == ERR_FAULT && !strncmp(last_put, "ARTICLE", 7)) || (respcode == ERR_FAULT && !strcmp(last_put, "POST")) || (respcode == ERR_FAULT && !strcmp(last_put, "LIST")) || respcode == ERR_GOODBYE || respcode == OK_GOODBYE) && strcmp(last_put, "QUIT")) {
+	if (last_put[0] != '\0' && ((respcode == ERR_FAULT && (!strncmp(last_put, "ARTICLE", 7) || !strcmp(last_put, "POST") || !strcmp(last_put, "LIST"))) || respcode == ERR_GOODBYE || respcode == OK_GOODBYE) && strcmp(last_put, "QUIT")) {
 		/*
 		 * Maybe server timed out.
 		 * If so, retrying will force a reconnect.
@@ -1758,10 +1765,10 @@ get_only_respcode(
 		put_server(last_put);
 		ptr = tin_fgets(FAKE_NNTP_FP, FALSE);
 
-		if (tin_errno) {
+		if (tin_errno || ptr == NULL) {
 #	ifdef DEBUG
 			if (debug & DEBUG_NNTP)
-				debug_print_file("NNTP", "<<<%sError: tin_errno <> 0", logtime());
+				debug_print_file("NNTP", "<<<%sError: tin_errno<>0 or ptr==NULL in get_only_respcode(retry)", logtime());
 #	endif /* DEBUG */
 			return -1;
 		}
@@ -1771,9 +1778,11 @@ get_only_respcode(
 			debug_print_file("NNTP", "<<<%s%s", logtime(), ptr);
 #	endif /* DEBUG */
 		respcode = (int) strtol(ptr, &end, 10);
+		if (end == ptr) /* no leading numbers in response */
+			respcode = -1;
 		DEBUG_IO((stderr, "get_only_respcode(%d)\n", respcode));
 	}
-	if (message != NULL && mlen > 1)		/* Pass out the rest of the text */
+	if (message != NULL && mlen > 1 && *end != '\0')		/* Pass out the rest of the text */
 		my_strncpy(message, ++end, mlen - 1);
 
 	return respcode;
@@ -1815,13 +1824,16 @@ get_respcode(
 #	endif /* DEBUG */
 		STRCPY(savebuf, last_put);
 
-		if (!authenticate(nntp_server, userid, FALSE)) {
-			error_message(2, _(txt_auth_failed), nntp_caps.type == CAPABILITIES ? ERR_AUTHFAIL : ERR_ACCESS);
-			tin_done(EXIT_FAILURE);
-		}
+		if (!authenticate(nntp_server, userid, FALSE))
+			tin_done(EXIT_FAILURE, _(txt_auth_failed), nntp_caps.type == CAPABILITIES ? ERR_AUTHFAIL : ERR_ACCESS);
+
 		if (nntp_caps.type == CAPABILITIES) {
 			check_extensions();
 			can_post = nntp_caps.post && !force_no_post;
+		} else {
+			put_server("MODE READER");
+			if (get_only_respcode(message, mlen) == OK_CANPOST)
+				can_post = TRUE && !force_no_post;
 		}
 		if (curr_group != NULL) {
 			DEBUG_IO((stderr, _("Rejoin current group\n")));
@@ -1844,14 +1856,20 @@ get_respcode(
 			if (debug & DEBUG_NNTP)
 				debug_print_file("NNTP", "<<<%sError: tin_errno <> 0", logtime());
 #	endif /* DEBUG */
-				return -1;
+			return -1;
 		}
 
 #	ifdef DEBUG
 		if (debug & DEBUG_NNTP)
 			debug_print_file("NNTP", "<<<%s%s", logtime(), ptr);
 #	endif /* DEBUG */
+		if (ptr == NULL)
+			return -1;
+
 		respcode = (int) strtol(ptr, &end, 10);
+		if (end == ptr)	/* no leading numbers in response */
+			return -1;
+
 		if (message != NULL && mlen > 1)				/* Pass out the rest of the text */
 			strncpy(message, end, mlen - 1);
 	}
@@ -1961,8 +1979,7 @@ list_motd(
 					debug_print_file("NNTP", "<<<%s%s", logtime(), ptr);
 #	endif /* DEBUG */
 				/*
-				 * according to draft-elie-nntp-list-additions-00.txt 2.4.2
-				 * the MOTD is in UTF-8
+				 * RFC 6048 2.5.2 "The information MUST be in UTF-8"
 				 *
 				 * TODO: - store a hash value of the entire motd in the server-rc
 				 *         and only if it differs from the old value display the
