@@ -3,10 +3,10 @@
  *  Module    : save.c
  *  Author    : I. Lea & R. Skrenta
  *  Created   : 1991-04-01
- *  Updated   : 2011-04-17
+ *  Updated   : 2016-02-25
  *  Notes     :
  *
- * Copyright (c) 1991-2011 Iain Lea <iain@bricbrac.de>, Rich Skrenta <skrenta@pbm.com>
+ * Copyright (c) 1991-2016 Iain Lea <iain@bricbrac.de>, Rich Skrenta <skrenta@pbm.com>
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -49,16 +49,21 @@
 #endif /* HAVE_UUDEVIEW_H */
 
 #ifndef HAVE_LIBUU
-#undef OFF
-enum state { INITIAL, MIDDLE, OFF, END };
+#	undef OFF
+	enum state {
+		INITIAL,
+		MIDDLE,
+		OFF,
+		END
+	};
 #endif /* !HAVE_LIBUU */
 
 enum action {
 	VIEW,
 	SAVE,
-	SAVE_TAGGED
+	SAVE_TAGGED,
+	PIPE_RAW
 #ifndef DONT_HAVE_PIPING
-	, PIPE_RAW
 	, PIPE
 #endif /* !DONT_HAVE_PIPING */
 };
@@ -88,8 +93,8 @@ static void generate_filename(char *buf, int buflen, const char *suffix);
 #endif /* !DONT_HAVE_PIPING */
 static void post_process_uud(void);
 static void post_process_sh(void);
-static void process_part(t_part *part,	FILE *rawfp, FILE *outfile, const char *savepath, enum action what);
-static void process_parts(t_part *part,	FILE *rawfp, enum action what);
+static void process_part(t_part *part, t_openartinfo *art, FILE *outfile, const char *savepath, enum action what);
+static void process_parts(t_part *part, t_openartinfo *art, enum action what);
 static void show_attachment_page(void);
 static void start_viewer(t_part *part, const char *path);
 static void tag_pattern(void);
@@ -138,6 +143,7 @@ check_start_save_any_news(
 	int i, j;
 	int art_count, hot_count;
 	int saved_arts = 0;					/* Total # saved arts */
+	struct t_article *art;
 	struct t_group *group;
 	t_bool log_opened = TRUE;
 	t_bool print_first = verbose;
@@ -202,7 +208,7 @@ check_start_save_any_news(
 			snprintf(buf, sizeof(buf), _(txt_saved_groupname), group->name);
 			fprintf(fp_log, "%s", buf);
 			if (verbose)
-				wait_message(0, buf);
+				wait_message(0, "%s", buf);
 
 			if (function == SAVE_ANY_NEWS) {
 				char tmp[PATH_LEN];
@@ -216,8 +222,14 @@ check_start_save_any_news(
 			}
 		}
 
-		if (!index_group(group))
+		if (!index_group(group)) {
+			for_each_art(j) {
+				art = &arts[j];
+				FreeAndNull(art->refs);
+				FreeAndNull(art->msgid);
+			}
 			continue;
+		}
 
 		/*
 		 * For each article in this group...
@@ -251,7 +263,7 @@ check_start_save_any_news(
 						continue;
 
 					if (function == SAVE_ANY_NEWS) {
-						snprintf(buf, sizeof(buf), "%ld", arts[j].artnum);
+						snprintf(buf, sizeof(buf), "%"T_ARTNUM_PFMT, arts[j].artnum);
 						joinpath(savefile, sizeof(savefile), path, buf);
 					}
 
@@ -284,10 +296,10 @@ check_start_save_any_news(
 						fprintf(savefp, "Content-Transfer-Encoding: 8bit\n\n");
 					}
 
-					snprintf(buf, sizeof(buf), "[%5ld]  %s\n", arts[j].artnum, arts[j].subject);
+					snprintf(buf, sizeof(buf), "[%5"T_ARTNUM_PFMT"]  %s\n", arts[j].artnum, arts[j].subject);
 					fprintf(fp_log, "%s", buf);	/* buf may contain % */
 					if (verbose)
-						wait_message(0, buf);
+						wait_message(0, "%s", buf);
 
 					while ((line = tin_fgets(artfp, FALSE)) != NULL)
 						fprintf(savefp, "%s\n", line);		/* TODO: error handling */
@@ -344,7 +356,7 @@ check_start_save_any_news(
 					group_count, PLURAL(group_count, txt_group));
 			fprintf(fp_log, "%s", buf);
 			if (verbose)
-				wait_message(0, buf);
+				wait_message(0, "%s", buf);
 
 			if (log_opened) {
 				fclose(fp_log);
@@ -421,7 +433,7 @@ open_save_filename(
 	}
 
 	if ((fp = fopen(path, mode)) == NULL) {
-		perror_message("%s (%s)", _(txt_art_not_saved), path);
+		error_message(2, _(txt_cannot_open_for_saving), path);
 		return NULL;
 	}
 
@@ -589,26 +601,34 @@ t_bool
 create_path(
 	const char *path)
 {
-	char buf[PATH_LEN];
-	int i, j, len;
+	char *buf, *p;
 	struct stat st;
 
-	len = (int) strlen(path);
+	if (!strlen(path))
+		return FALSE;
 
-	for (i = 0, j = 0; i < len; i++, j++) {
-		buf[j] = path[i];
-		if (i + 1 < len && path[i + 1] == DIRSEP) {
-			buf[j + 1] = '\0';
-			if (stat(buf, &st) == -1) {
-				if (my_mkdir(buf, (mode_t) (S_IRWXU|S_IRUGO|S_IXUGO)) == -1) {
-					if (errno != EEXIST) {
-						perror_message(_(txt_cannot_create), buf);
-						return FALSE;
-					}
+	buf = my_strdup(path);
+	p = buf + 1;
+
+	if (!strlen(p)) {
+		free(buf);
+		return FALSE;
+	}
+
+	while ((p = strchr(p, DIRSEP)) != NULL) {
+		*p = '\0';
+		if (stat(buf, &st) == -1) {
+			if (my_mkdir(buf, (mode_t) (S_IRWXU|S_IRUGO|S_IXUGO)) == -1) {
+				if (errno != EEXIST) {
+					perror_message(_(txt_cannot_create), buf);
+					free(buf);
+					return FALSE;
 				}
 			}
 		}
+		*p++ = DIRSEP;
 	}
+	free(buf);
 	return TRUE;
 }
 
@@ -1075,10 +1095,9 @@ view_file(
 	/*
 	 * Needed for the mime-type processor
 	 */
-	part->params = my_malloc(sizeof(t_param));
+	part->params = new_params();
 	part->params->name = my_strdup("name");
 	part->params->value = my_strdup(file);
-	part->params->next = NULL;
 
 	start_viewer(part, path);
 	my_printf(cCRLF);
@@ -1222,7 +1241,7 @@ start_viewer(
 			fflush(stdout);
 		} else {
 			if (foo->description)
-				info_message(foo->description);
+				info_message("%s", foo->description);
 		}
 		invoke_cmd(foo->command);
 		if (foo->needsterminal) {
@@ -1259,6 +1278,7 @@ decode_save_one(
 	FILE *fp;
 	char buf[2048], buf2[2048];
 	char *savepath;
+	int count;
 	int i;
 
 	/*
@@ -1277,7 +1297,6 @@ decode_save_one(
 	 * Decode/save the attachment
 	 */
 	if ((fp = open_save_filename(savepath, FALSE)) == NULL) {
-		error_message(2, _(txt_cannot_open_for_saving), savepath);
 		free(savepath);
 		return FALSE;
 	}
@@ -1296,8 +1315,6 @@ decode_save_one(
 			break;
 
 		switch (part->encoding) {
-			int count;
-
 			case ENCODING_QP:
 			case ENCODING_BASE64:
 				count = mmdecode(buf, part->encoding == ENCODING_QP ? 'q' : 'b', '\0', buf2);
@@ -1358,7 +1375,11 @@ decode_save_one(
 }
 
 
-enum match { NO, MATCH, NOTMATCH };
+enum match {
+	NO,
+	MATCH,
+	NOTMATCH
+};
 
 /*
  * Match a single type/subtype Content pair
@@ -1443,7 +1464,11 @@ check_save_mime_type(
 
 	ptr = my_strdup(mime_types);
 
-	pair = strtok(ptr, ",");
+	if ((pair = strtok(ptr, ",")) == NULL) {
+		free(ptr);
+		return FALSE;
+	}
+
 	retcode = match_content_type(part, pair);
 
 	while ((pair = strtok(NULL, ",")) != NULL) {
@@ -1521,6 +1546,7 @@ show_attachment_page(
 
 	signal_context = cAttachment;
 	currmenu = &attmenu;
+	mark_offset = 0;
 
 	if (attmenu.curr < 0)
 		attmenu.curr = 0;
@@ -1662,7 +1688,7 @@ attachment_page(
 			case ATTACHMENT_SAVE:
 				if (attmenu.max) {
 					part = get_part(attmenu.curr);
-					process_parts(part, art->raw, num_of_tagged_parts ? SAVE_TAGGED : SAVE);
+					process_parts(part, art, num_of_tagged_parts ? SAVE_TAGGED : SAVE);
 					show_attachment_page();
 				}
 				break;
@@ -1670,7 +1696,7 @@ attachment_page(
 			case ATTACHMENT_SELECT:
 				if (attmenu.max) {
 					part = get_part(attmenu.curr);
-					process_parts(part, art->raw, VIEW);
+					process_parts(part, art, VIEW);
 					show_attachment_page();
 				}
 				break;
@@ -1732,7 +1758,7 @@ attachment_page(
 			case GLOBAL_PIPE:
 				if (attmenu.max) {
 					part = get_part(attmenu.curr);
-					process_parts(part, art->raw, func == GLOBAL_PIPE ? PIPE_RAW : PIPE);
+					process_parts(part, art, func == GLOBAL_PIPE ? PIPE_RAW : PIPE);
 					show_attachment_page();
 				}
 				break;
@@ -1793,8 +1819,7 @@ build_attachment_line(
 	char buf[BUFSIZ];
 	char buf2[BUFSIZ];
 	char *tree = NULL;
-	int len, namelen, tagged;
-	int treelen = 0;
+	int len, namelen, tagged, treelen;
 	t_part *part;
 
 #ifdef USE_CURSES
@@ -1822,7 +1847,11 @@ build_attachment_line(
 
 	charset = get_param(part->params, "charset");
 	snprintf(buf2, sizeof(buf2), _(txt_attachment_lines), part->line_count);
-	snprintf(buf, sizeof(buf), "  %s/%s, %s, %s%s%s", content_types[part->type], part->subtype, content_encodings[part->encoding], charset ? charset : "", charset ? ", " : "", buf2);
+	/* TODO: make the layout configurable? */
+	if (!strcmp(content_types[part->type], "text"))
+		snprintf(buf, sizeof(buf), "  %s/%s, %s, %s%s%s", content_types[part->type], part->subtype, content_encodings[part->encoding], charset ? charset : "", charset ? ", " : "", buf2);
+	else
+		snprintf(buf, sizeof(buf), "  %s/%s, %s, %s", content_types[part->type], part->subtype, content_encodings[part->encoding], buf2);
 	if (part->depth > 0) {
 		treelen = cCOLS - 13 - info_len - namelen;
 		tree = build_tree(part->depth, treelen, i);
@@ -2154,11 +2183,11 @@ free_part_list(
 static void
 process_parts(
 	t_part *part,
-	FILE *rawfp,
+	t_openartinfo *art,
 	enum action what)
 {
 	FILE *fp;
-	char *savepath, *tmppath;
+	char *savepath = NULL, *tmppath;
 	int i, saved_parts = 0;
 	t_partl *lptr;
 
@@ -2173,11 +2202,10 @@ process_parts(
 							return;
 
 						if ((fp = open_save_filename(savepath, FALSE)) == NULL) {
-							error_message(2, _(txt_cannot_open_for_saving), savepath);
 							free(savepath);
 							return;
 						}
-						process_part(lptr->part, rawfp, fp, NULL, SAVE);
+						process_part(lptr->part, art, fp, NULL, SAVE);
 						free(savepath);
 						++saved_parts;
 					}
@@ -2197,11 +2225,10 @@ process_parts(
 				free(tmppath);
 			}
 			if ((fp = open_save_filename(savepath, FALSE)) == NULL) {
-				error_message(2, _(txt_cannot_open_for_saving), savepath);
 				free(savepath);
 				return;
 			}
-			process_part(part, rawfp, fp, savepath, what);
+			process_part(part, art, fp, savepath, what);
 			break;
 	}
 	switch (what) {
@@ -2231,35 +2258,75 @@ process_parts(
 static void
 process_part(
 	t_part *part,
-	FILE *rawfp,
+	t_openartinfo *art,
 	FILE *outfile,
 	const char *savepath,
 	enum action what)
 {
+	FILE *infile = NULL;
 	char buf[2048], buf2[2048];
-	int i;
+	int count;
+	int i, line_count;
+#ifdef CHARSET_CONVERSION
+	char *conv_buf;
+	const char *network_charset;
+	size_t line_len;
+#endif /* CHARSET_CONVERSION */
+
+	/*
+	 * uuencoded parts must be read from the cooked article,
+	 * otherwise they might be additionally encoded with b64 or qp
+	 */
+	if (part->encoding == ENCODING_UUE)
+		infile = art->cooked;
+	else
+		infile = art->raw;
 
 	if (what != PIPE_RAW && part->encoding == ENCODING_BASE64)
 		mmdecode(NULL, 'b', 0, NULL);				/* flush */
 
-	fseek(rawfp, part->offset, SEEK_SET);
+	fseek(infile, part->offset, SEEK_SET);
 
-	for (i = 0; i < part->line_count; i++) {
-		if ((fgets(buf, sizeof(buf), rawfp)) == NULL)
+	line_count = part->line_count;
+
+	for (i = 0; i < line_count; i++) {
+		if ((fgets(buf, sizeof(buf), infile)) == NULL)
 			break;
 
 		/* This should catch cases where people illegally append text etc */
 		if (buf[0] == '\0')
 			break;
 
+		/*
+		 * page.c:new_uue() sets offset to the 'begin ...' line
+		 * -> skip over the first line in uuencoded parts
+		 */
+		if (part->encoding == ENCODING_UUE && i == 0) {
+			++line_count;
+			continue;
+		}
+
 		if (what != PIPE_RAW) {
 			switch (part->encoding) {
-				int count;
-
 				case ENCODING_QP:
 				case ENCODING_BASE64:
-					count = mmdecode(buf, part->encoding == ENCODING_QP ? 'q' : 'b', '\0', buf2);
-					fwrite(buf2, count, 1, outfile);
+#ifdef CHARSET_CONVERSION
+					memset(buf2, '\0', sizeof(buf2));
+#endif /* CHARSET_CONVERSION */
+					if ((count = mmdecode(buf, part->encoding == ENCODING_QP ? 'q' : 'b', '\0', buf2)) > 0) {
+#ifdef CHARSET_CONVERSION
+						if (what != SAVE && what != SAVE_TAGGED && !strncmp(content_types[part->type], "text", 4)) {
+							line_len = count;
+							conv_buf = my_strdup(buf2);
+							network_charset = get_param(part->params, "charset");
+							process_charsets(&conv_buf, &line_len, network_charset ? network_charset : "US-ASCII", tinrc.mm_local_charset, FALSE);
+							strncpy(buf2, conv_buf, sizeof(buf2) - 1);
+							count = strlen(buf2);
+							free(conv_buf);
+						}
+#endif /* CHARSET_CONVERSION */
+						fwrite(buf2, count, 1, outfile);
+					}
 					break;
 
 				case ENCODING_UUE:
@@ -2272,6 +2339,16 @@ process_part(
 					break;
 
 				default:
+#ifdef CHARSET_CONVERSION
+						if (what != SAVE && what != SAVE_TAGGED && !strncmp(content_types[part->type], "text", 4)) {
+							conv_buf = my_strdup(buf);
+							line_len = strlen(conv_buf);
+							network_charset = get_param(part->params, "charset");
+							process_charsets(&conv_buf, &line_len, network_charset ? network_charset : "US-ASCII", tinrc.mm_local_charset, FALSE);
+							strncpy(buf, conv_buf, sizeof(buf) - 1);
+							free(conv_buf);
+						}
+#endif /* CHARSET_CONVERSION */
 					fputs(buf, outfile);
 			}
 		} else
@@ -2303,7 +2380,7 @@ static void
 pipe_part(
 	const char *savepath)
 {
-	FILE *fp, *pipe_fp = (FILE *) 0;
+	FILE *fp, *pipe_fp;
 	char *prompt;
 
 	prompt = fmt_string(_(txt_pipe_to_command), cCOLS - (strlen(_(txt_pipe_to_command)) + 30), tinrc.default_pipe_command);
